drop procedure if exists auto_plan_project;
drop procedure if exists auto_plan_calc_dayincrement;
drop procedure if exists test_dayincrement;
drop procedure if exists auto_plan_issue;
drop procedure if exists auto_plan_subissues;
-- drop trigger if exists issues_update_auto_plan;

update issues i set i.start_date=null, i.due_date=null where i.id not in (1, 9);

delimiter //

create procedure auto_plan_calc_dayincrement (
    in  thecurrentdate date,
    in  thespenthours float,
    in  theestimatedhours float,
    in  thecurrentuser int,
    out theresultdate date,
    out theresulthours int)
begin
    declare done int default 0;
    declare currentdate date default thecurrentdate;
    declare currentuser int default thecurrentuser;
    declare spenthours float default thespenthours;
    declare workload float default theestimatedhours;
    declare dayOfWeek int;
    declare hoursThatDay int;

    -- When there are no users assigned, the worktimes of admin are used.
    if (currentuser is null or currentuser = 0) then
        select id from users where login='admin' into currentuser;
    end if;

    if (currentdate is null) then
        set currentdate = now();
    end if;

    -- Don't allow null values - they will mess up calculations!
    if (workload is null) then
        set workload = 0;
    end if;
    
    if (spenthours is null) then
        set spenthours = 0;
    end if;
    
    repeat
        -- Get the day of week in the format we expect and use in users_worktime (1 = Monday...7 = Sunday)
        set dayOfWeek = DAYOFWEEK(currentdate);
        if dayOfWeek = 1 then
            set dayOfWeek = 7;
        else
            set dayOfWeek = (dayOfWeek - 1);
        end if;

        -- Select how many hours we can work at that weekday
        select w.hours_per_day from user_worktimes w where w.user_id=currentuser and w.day_of_week=dayOfWeek into hoursThatDay;
        if (hoursThatDay is null) then
            set hoursThatDay = 0;
        end if;
        
        -- select currentdate, hoursThatDay, workload, spenthours, (workload + spenthours)
        if (hoursThatDay >= (spenthours + workload)) then
            -- We have more hours left that day than we need. This means we found our due date!
            set theresulthours = hoursThatDay - (spenthours + workload);
            set theresultdate = currentdate;
            set done = 1;
        else
            -- We have less hours left that day than we need.
            -- The hours that we have left are subtracted from the workload, the date is incremented by one day
            -- and the ours already spent at the next day are set to 0.
            set currentdate = DATE_ADD(currentdate, INTERVAL 1 DAY);
            set workload = workload - (hoursThatDay - spenthours);
            set spenthours = 0;
        end if;
    until done != 0 end repeat;
end//

create procedure auto_plan_issue (in the_issue_id int,
                                  in thecurrentdate date, 
                                  in thespenthours float,
                                  out theduedate date)
begin
end//

create procedure auto_plan_subissues (in theparentissue int,
                                      in thecurrentuser int,
                                      in thecurrentdate date, 
                                      in thespenthours float,
                                      out theduedate date)
begin
    -- used to update related issues
    declare predecessorduedate date;
    declare discarddate date;
    declare issue_id int;
    declare v_parent_id int;
    declare user_id int;
    declare sub_done int default 0;
    declare sub_issues cursor for select i.id from issues i where
        (i.parent_id=theparentissue or
        exists (select r.* from issue_relations r 
            where
                r.relation_type in ('blocks', 'precedes') and
                r.issue_from_id=theparentissue and
                r.issue_to_id=i.id)) and
        (i.assigned_to_id = thecurrentuser or (thecurrentuser is null and i.assigned_to_id is null));
    declare continue handler for not found set sub_done=1;

    open sub_issues;
    repeat
        fetch sub_issues into issue_id;
        if not sub_done then
            select max(i.due_date) from issues i
                where
                    i.id=null or
                    exists (select r.* from issue_relations r
                            where
                                r.relation_type in ('blocks', 'precedes') and
                                r.issue_from_id=i.id and
                                r.issue_to_id=8)
                        into
                            predecessorduedate;
            if (predecessorduedate is not null) then
                if (predecessorduedate > thecurrentdate) then
                    set thecurrentdate = predecessorduedate;
                end if;
            end if;

            call auto_plan_issue (issue_id, thecurrentdate, thespenthours, discarddate);
            set thespenthours = 0;
        end if;
    until sub_done end repeat;
    close sub_issues;
end//

drop procedure auto_plan_issue//

create procedure auto_plan_issue (in the_issue_id int,
                                  in thecurrentdate date,
                                  in thespenthours float,
                                  out theduedate date)
begin
    -- for calls to auto_plan_calc_dayincrement
    declare currentdate date default thecurrentdate;
    declare spenthours float default thespenthours;
    declare estimatedhours float;
    declare currentuser int;
    declare resultdate date;
    declare resulthours int;

    select "Planning issue ", the_issue_id;

    -- Select needed data from the issue
    select i.estimated_hours, i.assigned_to_id from issues i where i.id=the_issue_id into estimatedhours, currentuser;
    
    -- Calculate the due_date from the start_date and the estimated hours and override the start_date
    call auto_plan_calc_dayincrement (currentdate, spenthours, estimatedhours, currentuser, resultdate, resulthours);
    update issues i set i.start_date=currentdate, i.due_date=resultdate where i.id=the_issue_id;
    
    -- Now select all following issues assigned to the same user
    select "Planning subissues for ", the_issue_id;
    call auto_plan_subissues (the_issue_id, currentuser, resultdate, resulthours, resultdate);
    set theduedate = resultdate;
    select "Planned subissues for ", the_issue_id;
end//

create procedure auto_plan_project (in the_project_id int)
begin
    declare issue_id int;
    declare user_id int;
    declare currentdate date;
    declare done int default 0;
    -- Select 'root' issues. These are automatically planned issues that do not have a predecessor and are not blocked by other issues
    declare root_issues cursor for select i.id, i.assigned_to_id, i.start_date from issues i
        where 
            i.project_id=the_project_id and 
            exists (select f.* from custom_fields f 
                left join
                    custom_values v on v.custom_field_id=f.id
                where 
                    f.type='IssueCustomField' and 
                    f.name='Automatisch planen' and
                    v.customized_id=i.id and
                    v.value='1') and
            i.parent_id is null and
            not exists (
                select r.* from issue_relations r 
                    where 
                        r.relation_type in ('blocks', 'precedes') and 
                        r.issue_to_id=i.id);
    declare continue handler for not found set done=1;

    -- Set all non-root, auto planned issues' due date to null, because they are being planned now.
    -- This is required because otherwise old, non-updated due dates could be used to calculate the minimum start
    -- date for sub-issues that have multiple predecessors or blockers.
    update issues i set i.due_date=null
        where 
            i.project_id=the_project_id and 
            exists (select f.* from custom_fields f 
                left join
                    custom_values v on v.custom_field_id=f.id
                where 
                    f.type='IssueCustomField' and 
                    f.name='Automatisch planen' and
                    v.customized_id=i.id and
                    v.value='1') and
            (i.parent_id is not null or
            exists (
                select r.* from issue_relations r 
                    where 
                        r.relation_type in ('blocks', 'precedes') and 
                        r.issue_to_id=i.id));

    if (@auto_plan_disable_issue_trigger != 1 or @auto_plan_disable_issue_trigger is null) then
        set @auto_plan_disable_issue_trigger=1;
        
        open root_issues;
        repeat
            fetch root_issues into issue_id, user_id, currentdate;
            if not done then
                -- Now auto plan this root issue and all issues related to it
                call auto_plan_issue (issue_id, currentdate, 0, currentdate);
            end if;
        until done end repeat;
        close root_issues;
        set @auto_plan_disable_issue_trigger=null;
    end if;
end//

-- create trigger issues_update_auto_plan after update on issues
-- for each row
-- begin
--     call auto_plan_project(new.project_id);
-- end;//

create procedure test_dayincrement()
begin
    declare currentdate date default '2012-01-20';
    declare spenthours float default 5;
    declare estimatedhours float default 7;
    declare currentuser int default 3;
    declare resultdate date;
    declare resulthours int;
    call auto_plan_calc_dayincrement (currentdate, spenthours, estimatedhours, currentuser, resultdate, resulthours);
    select resultdate, resulthours;
end//

delimiter ;

set @auto_plan_disable_issue_trigger=null;

-- set @auto_plan_disable_issue_trigger=1;
-- update issues set due_date=null;

call auto_plan_project(1);
